
# An HttpProtocol is created when a connection is received and is deleted when
# the connection is closed.
#
# As soon as it is created we start accepting data.  When a complete request is
# received we create a Request object and launch it on its own using
# asyncio.async.  When the request completes, it must call back into the
# protocol so it can continue parsing the next request.  If your handler never
# calls a completing method (e.g. request.complete_with_content) then a
# connection timeout (a TCP/IP timeout or our background reaper) will eventually
# clean up, but this is not ideal.
#
# I'm trying to keep the logic in Request to make it easier to make it generic
# and subclassable later.  (For now, I'm hardcoding our application Logic there
# though.)

# TODO: Reaper
# TODO: Handle connection close

import re, logging
from asyncio import Protocol, coroutine, async

from . import errors
from .errors import HttpError
from . import routing
from .contexts import Context
from .requests import Request
from .lowerdict import LowerDict
from .middleware import middleware

logger = logging.getLogger('web')

MAX_HEADERS = 1024 * 6
RE_REQUEST_LINE  = re.compile(r'(GET|POST|PUT|DELETE) [ ]+ (\S+) [ ]+ HTTP/1.1 [ ]* \r\n', re.VERBOSE)

_STATE_IDLE             = 0
_STATE_READING_HEADERS  = 1
_STATE_READING_CONTENT  = 2
_STATE_HANDLING_REQUEST = 3

_STATE_NAMES = { v: k for (k,v) in globals().items() if k.startswith('_STATE_') }

class HttpProtocol(Protocol):

    _next_id = 1

    def __init__(self):
        Protocol.__init__(self)

        self._id = HttpProtocol._next_id
        HttpProtocol._next_id += 1

        self.ip = None
        # The IP4 address we are directly connected to.

        self.buffer = b''

        self.state = _STATE_READING_HEADERS

        self.transport = None
        # Set to the transport while we are connected.  Once we close the
        # connection or it is closed on us, we set this back to None.

        self.headers = None
        # Contains the method line and all headers (everything up to the first
        # blank line).

        self.request_length = None
        # The content length of the request.

        self.method  = None
        self.url     = None
        self.match   = None
        self.route   = None
        self.ctx = None # The current request we are handling

    def connection_made(self, transport):
        assert self.ip is None

        self.transport = transport

        peername = transport.get_extra_info('peername')
        self.debug('Connection from %s', peername)

        if isinstance(peername, tuple) and len(peername) == 2:
            # An IP4 address.
            self.ip = peername[0]

        self.state = _STATE_READING_HEADERS

    def debug(self, msg, *args, exc_info=None):
        logger.debug(msg, *args, exc_info=exc_info)

    def info(self, msg, *args, exc_info=None):
        logger.info(msg, *args, exc_info=exc_info)

    def error(self, msg, *args, exc_info=None):
        logger.error(msg, *args, exc_info=exc_info)

    def connection_lost(self, exc):
        self.debug('connection_lost')
        self.transport = None

    def data_received(self, data):
        # data: Bytes
        self.buffer += data
        self.debug('data_received')
        self._process_buffer()

    def _process_buffer(self):
        self.debug('_process_buffer')
        if self.state == _STATE_READING_HEADERS:
            parts = self.buffer.split(b'\r\n\r\n', 1)

            if len(parts[0]) > MAX_HEADERS:
                # (By putting the check here we check the length when the
                # end-of-headers has been found and when it hasn't.)
                self.error('Headers too large: length=%s', len(parts[0]))
                raise HttpError(431)

            if len(parts) == 1:
                # Haven't seen the end-of-headers yet.
                return

            self.buffer = parts[1]

            self.method, self.url, headers = self.parse_start_line(parts[0])
            self.headers = self.parse_headers(headers)

            self.request_length = int(self.headers.get('content-length', 0))

            self.state = _STATE_READING_CONTENT

        if self.state == _STATE_READING_CONTENT and len(self.buffer) >= self.request_length:
            # Remember, if multiple requests are pipelined there can be *more*
            # data in the buffer than the content length.

            body = self.buffer[:self.request_length]
            self.buffer = self.buffer[self.request_length:]

            self.state = _STATE_HANDLING_REQUEST
            self.handle_request(body)


    def parse_start_line(self, buffer):
        line, headers = buffer.split(b'\r\n', 1)

        tokens = line.split(None)
        # method url HTTP/1.1
        if len(tokens) != 3 or tokens[0] not in (b'GET', b"POST", b'PUT', b'DELETE'):
            self.error('Invalid start line: "%s"', line)
            raise HttpError(400)

        # TODO: Decode the URL

        method = str(tokens[0], encoding='ascii')
        url    = str(tokens[1], encoding='ascii')

        return method, url, headers


    def parse_headers(self, raw):
        raw = str(raw, encoding='ascii')

        headers = LowerDict()

        for line in raw.split('\r\n'):
            parts = line.split(':', 1)
            if len(parts) != 2 or not len(parts[0]) or ' ' in parts[0]:
                # The specification says we must reject the request if there is
                # space before the colon.  I don't know of any header namesthat
                # contain spaces, so we'll simply reject if there are any spaces
                # in the name at all.
                #
                # Note that we also do not accept continuation lines.  Either
                # there won't be a ':' or there will be a leading space, either
                # of which cause us to reject.  The HTTP 1.1 specification
                # recommends we reject continuation lines.
                self.error('Invalid header "%r".  headers="%r"', line, raw)
                raise HttpError(400)

            headers[parts[0].lower()] = parts[1].strip()

        return headers


    def handle_request(self, body):

        # It doesn't seem possible to use async (ensure_future) and pass
        # parameters.  I guess we have to first launch a function with no
        # parameters that then yields to the one we want.  I've put the handler
        # into the request.

        self.route, self.match = routing.find(self.method, self.url)

        ip = self.headers.get('X-Forwarded-For', self.ip)

        request = Request(self, self.method, self.url, self.headers, body)
        self.ctx = Context(self.route, request, ip)

        async(self._handle_request_coroutine())


    @coroutine
    def _handle_request_coroutine(self):
        ctx = self.ctx

        try:
            for m in middleware:
                # The start method should not return anything.  Therefore, if anything came
                # back, it must be a generator or coroutine.  (We cannot use
                # inspect.isawaitable, et. al. which seems to have stopped working in 3.5.2.
                # It seems our methods with @asyncio.coroutine get marked as ITERABLE_COROUTINE
                # but the function does not derive from types.GeneratorType, so isawaitable
                # returns False.)
                gen = m.start(ctx)
                if gen:
                    yield from gen

            if not self.route:
                raise HttpError(404, ctx.url, method=ctx.method)

            ctx.response.body = yield from self.route(self.match, ctx)

        except HttpError as ex:
            headers = []

            if (301 <= ex.code <= 308):
                self.info('HTTP redirect %s %s', ex.code, ex.url)
                assert ex.url, 'Forgot to set URL in error for redirect? %r' % ex
                headers.append(b'Location: ' + ex.url.encode('ascii'))
            else:
                self.error('HTTP error %s %s url=%s%s', ex.code, str(ex),
                        (ex.method and (ex.method + ' ') or ''), ctx.url)

            self.complete(ex.code, headers=headers)

        except:
            self.error('Unhandled error in %s', self.route, exc_info=True)
            self.complete(500)

        for m in reversed(middleware):
            gen = m.complete(ctx)
            if gen:
                yield from gen

        try:
            ctx.response._send(ctx, self.transport)
        except:
            # An error happened while we were sending.  (Try to do as little as
            # possible in _send so there is less chance of getting an exception
            # there.)  At this point our best bet is to abort.
            logger.error('An error occurred while trying to send: %r', ctx, exc_info=True)

        # Reset everything
        self.state = _STATE_READING_HEADERS

        self.request_length = None
        self.method  = None
        self.url     = None
        self.headers = None
        self.route   = None
        self.match   = None # Can we put this into the route or request?
        self.ctx     = None

        if self.buffer:
            self._process_buffer()

    def eof_received(self):
        pass

    def complete(self, code, headers=None, body=None):
        """
        headers
        : An optional sequence of headers as byte strings.  (This is not a dictionary!)
          Do not supply the content-length header.
        """
        if not self.transport:
            # Already disconnected.
            return

        assert isinstance(code, int)
        assert code != 204 or body is None, '204 cannot have body'

        self.ctx.response.status = code
        # REVIEW: Is this required or is the response status only useful for going the other
        # way and setting the code passed here?

        response = [
            b'HTTP/1.1 ' + bytes(str(code), 'ascii') + b' ' + errors.STATUSES[code],
            b'Content-Length: ' + (b'0' if not body else bytes(str(len(body)), 'ascii')),
        ]

        if headers:
            assert all(isinstance(h, bytes) for h in headers), repr(headers)
            # We provide content-length
            assert all(b'content-length:' not in h.lower() for h in headers)
            response.extend(headers)

        response.append(b'')

        if body:
            assert isinstance(body, bytes), 'Forgot to convert body to bytes: {!r}'.format(body)
            response.append(body)

        response = b'\r\n'.join(response)

        self.transport.write(response)

        # Don't reset flags yet.  Wait until the handler returns in
        # _handle_request_coroutine



"""
Provides the Response object.
"""

import os, logging, gzip
from . import errors
from .staticfiles import File
from .lowerdict import LowerDict

# TODO:: Need to enable way to make this secure.
HTTP_COOKIE_PATH   = '/'
HTTP_COOKIE_SECURE = ''

logger = logging.getLogger('servant')

def itob(n):
    """
    Int to bytes
    """
    return bytes(str(n), 'ascii')


class Response:
    """
    Encapsulates the response to send to the client.

    A URL handler does not have to interact with this object - it can simply
    return the value to send.  However, it may set the status, headers, cookies,
    or the body.
    """

    _ohook = None
    # The JSON encoding hook (json.dumps(default=xx)).  Set this using
    # configuration.config(encode_hook).

    def __init__(self):
        self.status  = None
        self.headers = LowerDict()
        self.cookies = {}
        self.body    = None

    def set_cookie(self, name, value, expires=None, http_only=True, secure=True):
        """
        Sets a response cookie
        """
        assert isinstance(name, str)
        assert isinstance(value, str)
        parts = [ value ]
        if expires:
            parts.append('Expires={}'.format(expires.strftime('%a, %d-%b-%Y %T GMT')))
        if http_only:
            parts.append('HttpOnly')
        if secure:
            parts.append('Secure')
        self.cookies[name] = '; '.join(parts)

    def delete_cookie(self, name):
        """
        Deletes a cookie by setting the expiration date to a time in the past.
        """
        assert isinstance(name, str)
        self.cookies[name] = 'deleted; Expires=Thu, 01-Jan-1970 00:00:01 GMT; HttpOnly'

    def _send(self, ctx, transport):
        """
        Called at the end of processing to send response to the browser.  Do not
        call this from a URL handler.
        """
        status = self.status
        body   = self.body

        if type(body) not in (type(None), bytes):
            # (In development, assert which will raise an exception.  If it gets out of
            # development, log it and return an error to the browser.)
            logger.error('Response is not bytes: {} {!r}'.format(type(body), body))
            raise AssertionError('Response is not bytes: {} {!r}'.format(type(body), body))
            status = 500
            body   = None

        if not status:
            if body is None:
                status = 204
            else:
                status = 200

        if body is not None:
            self.headers['content-length'] = str(len(body))

        if __debug__:
            for key, val in self.headers.items():
                assert type(val) is str, 'Header %s value is not a string: val=%r type=%s' % (key, val, type(val))

        headers = '\r\n'.join('{}: {}'.format(k,v) for k,v in self.headers.items())

        parts = [
            b'HTTP/1.1 ', itob(status), b' ', errors.STATUSES[status], b'\r\n',
            headers.encode('utf8'),
        ]

        if self.cookies:
            cs = [ 'Set-Cookie: {}={}; path={}{}'.format(key, val, HTTP_COOKIE_PATH, HTTP_COOKIE_SECURE)
                   for (key, val) in self.cookies.items() ]
            parts.append(b'\r\n')
            parts.append('\r\n'.join(cs).encode('utf8'))

        parts.append(b'\r\n\r\n')

        if body:
            parts.append(body)

        transport.writelines(parts)




# Import the most common items to simplify URL handlers code.

from .connection import HttpProtocol
from .staticfiles import File
from .routing import route, get, post, put, delete
from .middleware import middleware
from .middleware import register as register_middleware
from .errors import HttpError

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import gzip, json
from servant import File
from servant.responses import Response
from servant.middleware import Middleware

class ResponseMiddleware(Middleware):
    """
    Examines URL handler responses and converts them to bytes if necessary.
    """
    def complete(self, ctx):

        response = ctx.response

        if response.status:
            return

        body = response.body

        if body is None:
            response.status = 204
            return

        if isinstance(body, dict) or isinstance(body, list):
            response.status = 200
            response.body   = json.dumps(body, default=Response._ohook).encode('utf8')

            if len(response.body) > 200:
                response.headers['content-encoding'] = 'gzip'
                response.body = gzip.compress(response.body)
                
            response.headers['content-type']   = 'application/json'
            response.headers['content-length'] = len(response.body)
            return

        if isinstance(body, File):
            if body.etag and body.etag == ctx.request.headers.get('if-none-match', None):
                response.status = 304
                response.body   = None
                return

            if body.etag:
                response.headers['etag'] = body.etag

            if body.compressed:
                response.headers['content-encoding'] = 'gzip'

            response.status = 200
            response.body   = body.content
            response.headers['content-type']   = body.mimetype
            response.headers['content-length'] = len(response.body)

            if body.cache_control:
                response.headers['cache-control'] = body.cache_control

            return

        if not isinstance(body, bytes):
            raise Exception('Response is not bytes: ctx=%s resp=%s' % (ctx, body))

        response.status = 200



"""
Provides the Request object.
"""

import json, re
from cgi import parse_qs
from urllib.parse import urlsplit
from cookies import Cookies

ct_appjson = re.compile('^[^;]+/json(;|$)')

class Request:
    """
    Encapsulates the request information in a usable form.

    id
      An internal counter that may be useful for troubleshooting.  It is written to the log to
      help collate log lines.  Note that this value wraps at 2 billion and is not unique over
      time.  It should be plenty for logging.


    headers
      A dictionary containing the request headers.  The dictionary will
      automatically lowercase keys when inserted or used for lookup.

    cookies
      A dictionary (case-sensitive) mapping cookie names to values.

    form
      A dictionary of variables from either the query string if a GET
      or from form or JSON data if it is a POST.  These are preparsed
      so they can be passed to the route in parameters.
    """
    _next_id = 1

    _ohook = None
    # The json.loads object_hook.  Set this using configuration.config(decode_hook).

    def __init__(self, cnxn, method, url, headers, body):
        self.cnxn    = cnxn
        self.method  = method
        self.url     = url
        self.headers = headers
        self.body    = body

        # A counter to help troubleshoot.
        self.id = Request._next_id
        Request._next_id = (Request._next_id + 1) % 2^32-1

        self.form = self.parse_form()

        c = headers.get('cookie', None)
        if c:
            self.cookies = Cookies.from_request(c)
        else:
            self.cookies = Cookies()

    def __repr__(self):
        # Format for debugging, not normal logging.
        return '{}/{} {} body={} bytes'.format(self.id, self.method, self.url, len(self.body))

    def parse_form(self):
        """
        Stores variables into self.form so they can be passed to the URL handlers.

        POSTs will be parsed if the content is JSON or form encoded.  GETs will
        have their variables parsed.
        """
        if self.method == 'GET':
            return { key: val[0] for (key, val) in parse_qs(urlsplit(self.url)[3]).items() }

        if self.method in ('POST', 'PUT'):
            ct = self.headers.get('content-type') or ''
            if ct == 'application/x-www-form-urlencoded':
                return { key: val[0] for (key, val) in parse_qs(self.body.decode('utf8'), True).items() }

            # TODO: Look for charset, etc.
            if ct_appjson.match(ct):
                s = self.body.decode('UTF-8')
                return json.loads(s, object_hook=Request._ohook)

        return None




def config(encode_hook=None, decode_hook=None):
    """
    Provides high-level configuration for the entire package.  It is not required; it is
    designed to allow customization.

    encode_hook
      A function that accepts a URL handler response as either a list or dictionary and JSON
      encodes it to a UTF-8 encoded string.  If not provided, json.dumps is used.

    decode_hook
      A JSON object hook (for json.loads(object_hook)) that is called for every object

      A function that accepts a Unicode JSON string and decodes it into a Python dictionary.
      If not provided, json.loads is used.

      A dictionary is required since it is used to populate URL handler arguments and we need
      to know the name of the argument each element should be assigned to.

    annotation_converter
      A function that post processes request variables (after the decode_hook, etc.) and
      optionally converts values based on Python argument annotations.

      This function is only called if the Route has annotations.  If so, the variables are
      passed as a mapping from argument name to value.

        func(route, map_arg_to_value) -> None

      If a value is to be converted, the new value should overwrite the old one in the map.
    """
    # Note: Push even if the value is None in case we're *restoring* the original value.  (I
    # don't see that being useful, but it would be expected if someone passed None on a second
    # call.)

    from .requests import Request
    Request._ohook = decode_hook

    from .responses import Response
    Response._ohook = encode_hook



"""
Provides the @route decorator and its implementation.

The decorator creates a DynamicRoute object, defined here, and stores it in the global
list of routes (`_routes`).  There is also a lookup function to find the
appropriate route for a URL.
"""

# REVIEW: We could implement static files as middleware, but I don't
# want to pay the performance penalty.  A single page app is rarely
# going to load a static file.  Also, if it is a big app it might be
# using a CDN or proxy and never load a static file

import re, inspect
from urllib.parse import urlparse
from asyncio import coroutine
from .middleware import middleware

_routes = []
# The global list of registered routes as DynamicRoute objects.


def register_route(r):
    for m in middleware:
        m.register_route(r)
    _routes.append(r)


def route(pattern, *, method=None, **kwargs):
    """
    The @route decorator used to register URL handlers.  The first parameter of
    the decorated function should be named "ctx".

    pattern
      The pattern for the URL the decorated function will handle.

      Wrap a component of the path in braces to mark it as a variable
      ("/file/{filename}").  The decorated function must take a parameter with
      this name.

    logger
      Optional Python logging.Logger instance.  If provided, the route
      parameters will be logged to it using logger.debug.
    """
    s = inspect.stack()
    logger = s[2].frame.f_globals.get('logger')

    assert method, 'Forgot to set "method" keyword in @route for %r.  You probably want to use @get or @post.' % pattern
    assert method in ('GET', 'POST', 'PUT', 'PATCH', 'DELETE'), 'invalid method: %r' % method

    def wrapper(func):
        # TODO: Convert to "==".
        assert not _find(pattern), 'URL "{}" registered twice: first={} second={}'.format(pattern, _find(pattern).funcname, func.__name__)

        r = DynamicRoute(pattern, func, kwargs, method=method, logger=logger)
        for m in middleware:
            m.register_route(r)

        _routes.append(r)
    return wrapper


def get(pattern, **kwargs):
    # You use "method" to set the method, but the point of having a function named "get" is to
    # set the method to "GET".  If you need another method, use @route instead.
    assert 'method' not in kwargs, 'Do not use "method" with get.  Use @route if you need another method.'
    kwargs['method'] = 'GET'
    return route(pattern, **kwargs)

def post(pattern, **kwargs):
    assert 'method' not in kwargs, 'Do not use "method" with post.  Use @route if you need another method.'
    kwargs['method'] = 'POST'
    return route(pattern, **kwargs)

def put(pattern, **kwargs):
    assert 'method' not in kwargs, 'Do not use "method" with put.  Use @route if you need another method.'
    kwargs['method'] = 'PUT'
    return route(pattern, **kwargs)

def delete(pattern, **kwargs):
    assert 'method' not in kwargs, 'Do not use "method" with delete.  Use @route if you need another method.'
    kwargs['method'] = 'DELETE'
    return route(pattern, **kwargs)


def _find(pattern):
    for r in _routes:
        if getattr(r, 'pattern', None) == pattern:
            return r
    return None

def find(method, url):
    """
    Given a URL, find the Route that handles it.

    If found, a tuple containing the Route object and the regular expression match used to
    match the pattern is returned.  If there are variables in the URL, the match will have a
    group for each in the order they are found in the URL.

    If not found, (None, None) is returned.
    """
    assert method in ('GET', 'POST', 'PUT', 'PATCH', 'DELETE'), 'Invalid method %r' % method

    url = urlparse(url)

    for r in _routes:
        if r.method != method:
            continue

        match = r.regexp.match(url.path)
        if match:
            return (r, match)

    return (None, None)


class Route:
    """
    The base class for routes.
    """
    def __init__(self, route_keywords=None, *, method, logger=None):
        self.route_keywords = route_keywords or {}
        self.method = method
        self.logger = logger


class DynamicRoute(Route):
    """
    A route that calls a user-defined function.  Instances of this are
    registered by the @route decorator.

    Variables can be created in the URL pattern by wrapping components
    in braces, such as "/static/js/{filename}".

    This object is callable like a function and will pick the arguments to the
    URL handler from the request (GET variables, JSON variables, etc.)
    """
    def __init__(self, pattern, func, route_keywords, *, method, logger=None):
        """
        pattern
          The URL regexp.

        func
          The URL handler callback.

        route_keywords
          A dictionary of keyword arguments passed to the @route decorator.
        """
        Route.__init__(self, route_keywords=route_keywords, logger=logger, method=method)

        self.pattern = pattern
        self._func = func
        self.func = coroutine(func)

        self.regexp = None

        self.urlvars = []
        # The names of variables to be parsed from the URL, in the order they
        # appear in the pattern.  There is a group for each in `self.regexp` so
        # all we have to do is get all of the groups.

        self.formvars = []
        # The names of variables we expect to find in vars variables or a JSON
        # object (depending on the content-type).  These are the URL handler
        # parameters that are not variables in the URL pattern.

        self.annotations = {}
        # The optional annotations describing the argument types for the URL handler function.
        #
        # This is the annotations from `inspect.getfullargspect`.  It is a mapping from
        # argument name to its type.  Only arguments that have an annotation are present.
        #
        # This is available to the functions reading URL and form variables to provide for
        # custom conversions, such as from an encoded Javascript date to a Python date,
        # datetime, or time.  (Javascript's Date always has date and time components, so by
        # annotating the server component, the system can remove the date or time as necessary
        # to get exactly the data type.)

        self.analyze_pattern()
        self.analyze_params()

    def analyze_params(self):
        """
        Analyzes the URL handlers parameters to determine which should come from URL
        variables and which are expected to be in the body.
        """
        spec = inspect.getfullargspec(self._func)

        if spec.annotations:
            self.annotations = spec.annotations

        assert spec.args and spec.args[0] == 'ctx', 'The first parameter is not `ctx`.  pattern={} callback={}'.format(self.pattern, self._func)

        args = set(spec.args[1:])
        self.formvars = args - set(self.urlvars)

    def analyze_pattern(self):
        """
        Creates a regular expression from the pattern for matching URLs.
        """
        # Split the URL by slashes and examine each part.  Each could be either
        # plan text (e.g. "static") or a variables (e.g. "{filename}").

        regexps  = []
        varnames = []

        assert self.pattern.startswith('/'), 'DynamicRoute patterns must start with "/": {!r}'.format(self.pattern)

        for part in self.pattern.rstrip('/').split('/'):
            if part.startswith('{') and part.endswith('}'):
                # variable
                varnames.append(part[1:-1])
                regexps.append('([^/]+)')
            else:
                # plain text
                regexps.append(re.escape(part))

        regexp = '^' + '/'.join(regexps) + '/?$'

        self.regexp = re.compile(regexp)
        self.urlvars = varnames

    @coroutine
    def __call__(self, match, ctx):
        """
        Calls the URL handler, passing any defined parameters.
        """
        args = {
            'ctx': ctx
        }

        if self.urlvars:
            args.update(zip(self.urlvars, match.groups()))

        if self.formvars and ctx.request.form:
            args.update({ name: ctx.request.form.get(name, None) for name in self.formvars })

        if self.logger:
            self.logger.debug('URL: %s params=%r', self.pattern, args)

        result = yield from self.func(**args)
        return result

    def __repr__(self):
        return 'DynamicRoute<{} {}>'.format(self.pattern, self._func)

    @property
    def funcname(self):
        return self._func.__name__




class LowerDict(dict):
    """
    A dictionary subclass that lowercases all keys.
    """
    def __setitem__(self, key, value):
        return dict.__setitem__(self, key.lower(), value)

    def __getitem__(self, key):
        return dict.__getitem__(self, key.lower())

    # REVIEW: Do we need to subclass `get` also or does __getitem__ handle that?




from . import Middleware
import logging
from time import perf_counter

logger = logging.getLogger('web')

class LoggingMiddleware(Middleware):
    def start(self, ctx):
        # The start and stop time should probably be part of the context anyway.
        ctx._log_timer_start = perf_counter()

    def complete(self, ctx):
        diff = perf_counter() - ctx._log_timer_start
        logger.info('%s status=%s time=%.7s seconds', ctx.url, ctx.response.status, diff)




# The only HTML file we serve that isn't a template, index.html, should have an
# etag and be checked each time.  Items inside all have a version as part of the
# filename, so reloading this one HTML file updates the site.
#
# All vendor files should have the version manually appended when adding to the
# project or upgrading versions.
#
# To ensure an incremental build that only updates one of the generated files
# works, the build scripts should always regenerate the index HTML too.

# WARNING
# -------
#
# Python's asyncio does not have an asynchronous way to read files because operating systems
# don't really support it.  Apparently *nix systems have the API (select with file descriptors)
# but they always report themselves ready and therefore end up blocking anyway.
#
# Python will soon (or may already) support an asynchronous sendfile, which would could try to
# use.
#
# To work around this, we'll simply cache the files in memory.  Due to reference counting,
# Python is usually very good with memory.  The library is really designed for single page
# applications where resources are cached at the browser for a year.  (Put your version number
# on the end!)

# TODO ITEMS
# ----------
#
# We have not implemented compression yet.  I don't believe there are any modern browsers that
# don't support gzip, so we should only support compression and hold the compressed version in
# memory.
#
# Add a way to register mimetypes.  (Or perhaps use a module that already has them?)

import re, gzip, os
from os.path import isdir, splitext, abspath, join, exists, isabs
from logging import getLogger
from .errors import HttpError
from asyncio import coroutine
from collections import namedtuple
from datetime import timedelta

from .routing import Route, register_route

logger = getLogger('static')

USE_CACHE = (os.environ.get('SERVANT_STATIC_CACHE', '1').lower() in ('1', 'true', 't'))
logger.debug('caching: %s', USE_CACHE)

Ext = namedtuple('Ext', 'mimetype compress')
map_ext_to_mime = {
    '.css'  : Ext('text/css', True),
    '.eot'  : Ext('application/vnd.ms-fontobject', True),
    '.gif'  : Ext('image/gif' , False),
    '.html' : Ext('text/html', True),
    '.ico'  : Ext('image/ico', False),
    '.jpe'  : Ext('image/jpeg', False),
    '.jpeg' : Ext('image/jpeg', False),
    '.jpg'  : Ext('image/jpeg', False),
    '.js'   : Ext('text/javascript', True),
    '.map'  : Ext('application/json', True),
    '.otf'  : Ext('application/x-font-otf', True),
    '.png'  : Ext('image/png', False),
    '.svg'  : Ext('image/svg+xml', True),
    '.ttf'  : Ext('application/x-font-ttf', True),
    '.woff' : Ext('application/font-woff', False),
    '.woff2': Ext('application/font-woff2', False)
}

re_dots = re.compile(r'(^|/)\.{1,2}(/|$)')
# A regular expression that matches ".." as a subdirectory (component) of a URL
# (../, /../, or /..).  This is not allowed.

routes = []
# Routes registered for static route prefixes, in the order they were registered.

class File:
    def __init__(self, relpath, mimetype, content, compressed, *, cache_control):
        self.relpath    = relpath
        self.mimetype   = mimetype
        self.content    = content
        self.compressed = compressed # True if gzipped
        self.etag       = None
        self.cache_control = cache_control


class StaticFileRoute(Route):
    """
    A route for serving static files from the static file cache.
    """
    def __init__(self, prefix, path, cache_control=None, route_keywords=None, name=None):
        """
        prefix
          The URL prefix.

        path
          The directory to load files from for URLs that start with the given prefix.

        cache_control
          The value to set for the cache-control header.  If not provided the
          header is not added.

        name
          An optional name for this route so it can be looked up later.
        """
        Route.__init__(self, route_keywords=route_keywords, method='GET')

        self.prefix = prefix
        self.path   = path
        self.name   = name

        self.cache_control = cache_control

        if prefix[-1] != '/':
            prefix += '/'

        self.regexp = self._prefix_to_regexp(prefix)

        self.cache = {}
        # Maps from relative path (from self.path) to a FileEntry object.  Used
        # when USE_CACHE is true.

    def __repr__(self):
        return 'StaticFileRoute<%s>' % self.prefix

    def _prefix_to_regexp(self, prefix):
        """
        Given a prefix being registered, convert it to a regular expression that
        matches the entire URL.  It should have a single group that matches
        everything after the prefix (and after the slash separating the prefix from
        the relative path).

        The special character '*' can be used to represent a subdirectory whose
        value is ignored.  This can be useful for ignoring version components of
        paths.  The star must be by itself and not mixed with letters.  For example
        "/static/*/img" and "/js/*" is accepted, but not "/static/x*/img".
        """
        # Escape all of the "normal" parts.  The only non-normal we support right
        # now is "*" which represents a subdirectory.
        parts = prefix.rstrip('/').split('/')
        parts = [ (re.escape(p) if p != '*' else '[^/]+') for p in parts ]

        regexp = '^' + '/'.join(parts) + '/(.+)$'

        return re.compile(regexp)

    def _getfile(self, relpath):
        """
        Returns an http.File object for the given self.path / relpath combination.
        """
        fqn = abspath(join(self.path, relpath))
        if not exists(fqn):
            logger.debug('Not found: url=%r fqn=%r', relpath, fqn)
            raise HttpError(404, relpath)

        if re_dots.search(relpath):
            # This means someone used ".." to try to move up out of the static directory.  This
            # very well may be a hack attempt.
            logger.error('SECURITY: Dangerous path in file download?  prefix=%s self.path=%s relpath=%s fqn=%s',
                        self.prefix, self.path, relpath, fqn)
            raise HttpError(404, relpath)

        ext = splitext(relpath)[1]
        if ext not in map_ext_to_mime:
            raise Exception('No mimetype for "{}" (from {!r})'.format(ext, relpath))

        content = open(fqn, 'rb').read()
        extinfo = map_ext_to_mime[ext]

        if extinfo.compress:
            content = gzip.compress(content)

        entry = File(relpath, extinfo.mimetype, content, extinfo.compress, cache_control=self.cache_control)

        return entry

    def get(self, relpath):
        """
        Returns the file, adding to the cache if necessary.
        """
        f = USE_CACHE and self.cache.get(relpath)
        if not f:
            f = self._getfile(relpath)

            if USE_CACHE:
                self.cache[relpath] = f

        return f

    @coroutine
    def __call__(self, match, ctx):
        relpath = match.group(1)
        return self.get(relpath)


def register_file_type(ext, mimetype=None, compress=None):
    map_ext_to_mime[ext] = Ext(mimetype=mimetype, compress=compress)


def serve_prefix(prefix, path, cache_control=None, name=None, **route_keywords):
    """
    Registers a URL prefix (e.g. "/images") with a directory.  Any URLs starting with this
    prefix will serve files from the given path or below.

    route_keywords
      Route keywords.  These are attached to the route for use by middleware.
    """
    assert isabs(path), 'static path {!r} for prefix {} is not absolute'.format(path, prefix)
    assert isdir(path), 'static path {!r} for prefix {} does not exist'.format(path, prefix)

    route = StaticFileRoute(prefix, path, cache_control=cache_control, name=name, route_keywords=route_keywords)
    register_route(route)

    routes.append(route)


def route_from_name(name):
    """
    Returns the route with the given name.  Returns None if not found.
    """
    for route in routes:
        if route.name and route.name == name:
            return route
    return None


def path_from_prefix(prefix):
    """
    Return the directory that files for the given URL prefix are being
    served from.
    """
    if prefix[-1] != '/':
        prefix += '/'

    for route in routes:
        if route.prefix == prefix:
            return route.path
    raise Exception('The prefix {!r} is not being served'.format(prefix))





# Remember, do not try to send user data back to the user in a message.  If they
# manage to sneak a '\r\n' into the data we'll have a header injection attack on
# our hands.

class HttpError(Exception):
    def __init__(self, code, url=None, method=None, message=None):
        assert isinstance(code, int)
        assert code in STATUSES, 'Invalid status {}'.format(code)
        assert code != 404 or url, 'Please provide the URL for 404s'
        Exception.__init__(self, message or str(STATUSES[code], 'utf8'))
        self.code = code
        self.method = method
        self.url    = url
        # This is used for both 404 and 3xx redirects.


STATUSES = {
    100: b"Continue",
    101: b"Switching Protocols",
    102: b"Processing",
    103: b"199",
    200: b"OK",
    201: b"Created",
    202: b"Accepted",
    203: b"Non-Authoritative Information",
    204: b"No Content",
    205: b"Reset Content",
    206: b"Partial Content",
    207: b"Multi-Status",
    208: b"Already Reported",
    226: b"IM Used",
    300: b"Multiple Choices",
    301: b"Moved Permanently",
    302: b"Found",
    303: b"See Other",
    304: b"Not Modified",
    305: b"Use Proxy",
    307: b"Temporary Redirect",
    308: b"Permanent Redirect",
    400: b"Bad Request",
    401: b"Unauthorized",
    402: b"Payment Required",
    403: b"Forbidden",
    404: b"Not Found",
    405: b"Method Not Allowed",
    406: b"Not Acceptable",
    407: b"Proxy Authentication Required",
    408: b"Request Timeout",
    409: b"Conflict",
    410: b"Gone",
    411: b"Length Required",
    412: b"Precondition Failed",
    413: b"Payload Too Large",
    414: b"URI Too Long",
    415: b"Unsupported Media Type",
    416: b"Range Not Satisfiable",
    417: b"Expectation Failed",
    418: b"I'm A Teapot",        # ;)
    420: b"Enhance Your Calm",
    421: b"Misdirected Request",
    422: b"Unprocessable Entity",
    423: b"Locked",
    424: b"Failed Dependency",
    426: b"Upgrade Required",
    428: b"Precondition Required",
    429: b"Too Many Requests",
    431: b"Request Header Fields Too Large",
    500: b"Internal Server Error",
    501: b"Not Implemented",
    502: b"Bad Gateway",
    503: b"Service Unavailable",
    504: b"Gateway Timeout",
    505: b"HTTP Version Not Supported",
    506: b"Variant Also Negotiates",
    507: b"Insufficient Storage",
    508: b"Loop Detected",
    510: b"Not Extended",
    511: b"Network Authentication Required",
}




from . import Middleware

class SecurityHeadersMiddlware(Middleware):
    def complete(self, ctx):
        # ctx.response.headers['Content-Security-Policy']   = "script-src 'self' 'unsafe-eval' ; style-src 'self' 'unsafe-inline'"
        ctx.response.headers['Strict-Transport-Security'] = 'max-age=8640000; includeSubDomains'
        ctx.response.headers['X-Frame-Options']           = 'DENY'
        ctx.response.headers['X-XSS-Protection']          = '1; mode=block'




middleware = []
# All middlware registered, in the order it was registered.

def register(m):
    middleware.append(m)


class Middleware:
    """
    A base class for middleware, primarily for documentation.  You do
    not need to call the base class methods from your subclass.
    """
    def register_route(self, route):
        pass

    def start(self, ctx):
        """
        Called at the beginning of request handling.

        This can be a normal function or a coroutine.
        """
        pass

    def complete(self, ctx):
        """
        Called at the end of request handling.

        This can be a normal function or a coroutine.
        """
        pass



"""
The context object stores all of the information in the request and the
response.  It provides convenience methods to "pass-through" to those objects.
"""

from .responses import Response

class Context:
    """
    A Context encapsulates the request and response objects and provides
    convenience methods.  Contexts are passed to middleware and to request
    handlers.

    A new Context is created for each request.

    Handlers normally just return the data they want sent to the browser, but
    they may also set the response status, headers, cookies, etc.
    """
    def __init__(self, route, request, ip):
        self.route = route
        # The route handling this request.

        self.request = request
        self.response = Response()

        self.ip = ip
        # The most likely address of the client.  If an X-Forwarded-For header
        # is present it is used.  Otherwise the address of the actual connection
        # is used.  Note that there is no way to ensure this is valid so be
        # careful when using it.

    @property
    def id(self):
        return self.request.id

    def __repr__(self):
        return repr(self.request)

    @property
    def headers(self):
        """
        The dictionary of request headers.  This dictionary always lowercases keys.

        This is a convenience property that returns ctx.request.headers.
        """
        return self.request.headers

    @property
    def method(self):
        return self.request.method

    @property
    def url(self):
        """
        The request URL.

        This is a convenience property that returns ctx.request.url.
        """
        return self.request.url

    # def create_session(self, user_id=None, auth_status=None, login=None, name=None):
    #     """
    #     Creates a new session-id and Session object and sets the 'sid' response
    #     cookie.
    #
    #     This should only be used when the user is logging in.  This cannot be
    #     used with a user that is already logged in.
    #     """
    #     assert self.session is None, 'There is already a session-id assigned to this context'
    #
    #     self.sid = genrandom(20)
    #     self.session = Session(id=self.sid, user_id=user_id, login=login, auth_status=auth_status,
    #                            name=name)
    #     self.set_cookie('sid', self.sid)

    @property
    def user_id(self):
        return self.session and self.session.user_id

    # def delete_session(self):
    #     """
    #     Marks the session for deleting and deletes the "sid" cookie.
    #
    #     The session middleware is response for actually deleting the session and
    #     the end of request processing.
    #     """
    #     if self.session:
    #         assert self._deleted_session is None, 'There is already a deleted session!'
    #         self._deleted_session = self.session
    #         self.session = None
    #
    #     self.delete_cookie('sid')
    #     self.sid = None

    def set_cookie(self, name, value, expires=None, http_only=True, secure=True):
        """
        Sets a response cookie.

        This is a convenience method that calls ctx.response.set_cookie.
        """
        self.response.set_cookie(name, value, expires=expires, http_only=http_only, secure=secure)

    def delete_cookie(self, name):
        """
        Deletes a cookie by settings its expiration date in the past.

        This is a convenience method that calls ctx.response.delete_cookie.
        """
        self.response.delete_cookie(name)





try:
    # I hate extra stuff, but pip looks at "install_requires" which is
    # only from setuptools.  Importing it will monkeypatch the dist
    # system and keep it from complaining about an unknown keyword.
    from setuptools import setup, Command
except:
    from distutils.core import Command, setup

import subprocess
from os.path import abspath, join, dirname, exists

import versioneer


class TestCommand(Command):

    # This is temporary until we get proper unit tests.  It just makes
    # it easier to fire up the "e1" test from the root directory.

    description = 'runs e1'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        cmd = 'python3 e1.py'
        p = subprocess.Popen(args=cmd, cwd=join(dirname(abspath(__file__)), 'example'), shell=True)
        p.wait()


class PyLintCommand(Command):
    description = "Runs pylint"

    user_options = [
        ('bad-functions=', 'f', 'functions that are reported (e.g. print)'),
        ('allow-unused=', 'u', 'a regexp of variables that can be unused'),
        ('errors-only', 'E', 'errors only'),
        ('no-reports', 'n', 'do not print reports'),
        ('ignore-docstring', 'd', 'ignore missing doc string message'),
        ('exclude-noisy', 'x', 'exclude docstring and fixme messages and disable reports')
    ]

    def initialize_options(self):
        self.errors_only = None
        self.no_reports = None
        self.bad_functions = None
        self.allow_unused = None
        self.ignore_docstring = None
        self.exclude_noisy = None

    def finalize_options(self):
        self.no_reports = bool(self.no_reports)
        self.ignore_docstring = bool(self.ignore_docstring)
        self.exclude_noisy = bool(self.exclude_noisy)

    def run(self):

        cmd = ['pylint']

        # The shared .pylintrc file is stored in this package.
        rcfile = join(dirname(abspath(__file__)), '.pylintrc')
        if exists(rcfile):
            cmd.append('--rcfile={}'.format(rcfile))

        if self.errors_only:
            cmd.append('-E')

        if self.no_reports or self.exclude_noisy:
            cmd.append('-rn')

        if self.bad_functions:
            cmd.append("--bad-functions={}".format(self.bad_functions))
        if self.allow_unused:
            cmd.append("--dummy-variables-rgx={}".format(self.allow_unused))

        if self.exclude_noisy:
            cmd.append("--disable=C0111,fixme")
        elif self.ignore_docstring:
            cmd.append("--disable=C0111")

        cmd.append('servant')

        if self.verbose >= 2:
            print('cmd:', ' '.join(cmd))

        return subprocess.call(cmd)


cmdclass = versioneer.get_cmdclass()
cmdclass.update({
    'lint': PyLintCommand,
    'test': TestCommand,
})

setup(
    name='Servant',
    description='Python 3 asyncio web server',
    author='Michael Kleehammer',
    author_email='michael@kleehammer.com',
    url='https://gitlab.com/mkleehammer/servant',
    version=versioneer.get_version(),
    packages=['servant', 'servant.middleware'],
    install_requires=['cookies'],
    cmdclass=cmdclass
)
